# from __future__ import print_function
#
# import random
# import unittest
#
# from EmbCompress import expand, compress
#
#
# class TestReadHus(unittest.TestCase):
#
#     def test_fake_compression(self):
#         for i in range(10):
#             s = random.randint(10, 100000)
#             test_bytes = bytearray(random.getrandbits(8) for _ in range(s))
#             compressed_bytes = compress(test_bytes)
#             uncompressed = bytearray(expand(compressed_bytes, len(test_bytes)))
#             self.assertEqual(test_bytes, uncompressed)



from __future__ import print_function

import unittest

from test.pattern_for_tests import *


class TestConverts(unittest.TestCase):

    def position_equals(self, stitches, j, k):
        self.assertEqual(stitches[j][:1], stitches[k][:1])

    def test_convert_tbf_to_u01(self):
        file1 = "convert_u01.tbf"
        file2 = "converted_tbf.u01"
        write_tbf(get_big_pattern(), file1)
        f_pattern = read_tbf(file1)
        write_u01(f_pattern, file2)
        t_pattern = read_u01(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(NEEDLE_SET), 16)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("tbf->u01: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_convert_tbf_to_csv(self):
        file1 = "convert_csv.tbf"
        file2 = "converted_tbf.csv"
        write_tbf(get_big_pattern(), file1)
        f_pattern = read_tbf(file1)
        write_csv(f_pattern, file2)
        t_pattern = read_csv(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(NEEDLE_SET), 16)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("tbf->csv: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_convert_tbf_to_exp(self):
        file1 = "convert_exp.tbf"
        file2 = "converted_tbf.exp"
        write_tbf(get_big_pattern(), file1)
        f_pattern = read_tbf(file1)
        write_exp(f_pattern, file2)
        t_pattern = read_exp(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("tbf->exp: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_convert_tbf_to_vp3(self):
        file1 = "convert_vp3.tbf"
        file2 = "converted_tbf.vp3"
        write_tbf(get_big_pattern(), file1)
        f_pattern = read_tbf(file1)
        write_vp3(f_pattern, file2)
        t_pattern = read_vp3(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("tbf->vp3: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_convert_tbf_to_jef(self):
        file1 = "convert_jef.tbf"
        file2 = "converted_tbf.jef"
        write_tbf(get_big_pattern(), file1)
        f_pattern = read_tbf(file1)
        write_jef(f_pattern, file2)
        t_pattern = read_jef(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("tbf->jef: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_convert_tbf_to_pec(self):
        file1 = "convert_pec.tbf"
        file2 = "converted_tbf.pec"
        write_tbf(get_big_pattern(), file1)
        f_pattern = read_tbf(file1)
        write_pec(f_pattern, file2)
        t_pattern = read_pec(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("tbf->pec: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_convert_tbf_to_tbf(self):
        file1 = "convert_tbf.tbf"
        file2 = "converted_tbf.tbf"
        write_tbf(get_big_pattern(), file1)
        f_pattern = read_tbf(file1)
        write_tbf(f_pattern, file2)
        t_pattern = read_tbf(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(NEEDLE_SET), 16)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("tbf->tbf: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_convert_tbf_to_dst(self):
        file1 = "convert_dst.tbf"
        file2 = "converted_tbf.dst"
        write_tbf(get_big_pattern(), file1)
        f_pattern = read_tbf(file1)
        write_dst(f_pattern, file2)
        t_pattern = read_dst(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("tbf->dst: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_convert_tbf_to_gcode(self):
        file1 = "convert_gcode.tbf"
        file2 = "converted_tbf.gcode"
        write_tbf(get_big_pattern(), file1)
        f_pattern = read_tbf(file1)
        write_gcode(f_pattern, file2)
        t_pattern = read_gcode(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("tbf->gcode: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_convert_tbf_to_pes(self):
        file1 = "convert_pes.tbf"
        file2 = "converted_tbf.pes"
        write_tbf(get_big_pattern(), file1)
        f_pattern = read_tbf(file1)
        write_pes(f_pattern, file2)
        t_pattern = read_pes(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("tbf->pes: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_convert_tbf_to_xxx(self):
        file1 = "convert_xxx.tbf"
        file2 = "converted_tbf.xxx"
        write_tbf(get_big_pattern(), file1)
        f_pattern = read_tbf(file1)
        write_xxx(f_pattern, file2)
        t_pattern = read_xxx(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("tbf->xxx: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_needles_convert_tbf_to_u01(self):
        file1 = "convert_u01.tbf"
        file2 = "converted_tbf.u01"
        write_tbf(get_shift_pattern_needles(), file1)
        f_pattern = read_tbf(file1)

        t = 8
        for x, y, cmd in f_pattern.stitches:
            if cmd & COMMAND_MASK == NEEDLE_SET:
                flag, thread, needle, order = decode_embroidery_command(cmd)
                self.assertEqual(needle, t)
                t -= 1
                if t <= 0:
                    t = 8

        write_u01(f_pattern, file2)
        t_pattern = read_u01(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(f_pattern.count_stitch_commands(NEEDLE_SET), 16)
        self.assertEqual(t_pattern.count_stitch_commands(NEEDLE_SET), 16)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        print("tbf->u01: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_tbf_flipped(self):
        """
        Test whether tbf is flipped on subsequent loads. This creates a pattern. Saves/loads, then save/loads.
        These patterns should be the same but a bug in 1.5.0 was flipping TBF patterns on write.
        """
        pattern = get_fractal_pattern()
        file1 = "fractal1.tbf"
        file2 = "fractal2.tbf"
        pattern.write(file1)
        f1 = read_tbf(file1)
        f1.write(file2)
        f2 = read_tbf(file2)
        for i in range(len(f1.stitches)):
            self.assertEqual(f1.stitches[i], f2.stitches[i])
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_needle_tbf_range(self):
        file1 = "test_range8.tbf"
        file2 = "test_range8.ct0"
        pattern = EmbPattern()
        pattern.metadata("name", "colorswitch8")
        pattern += "red"
        pattern += "blue"
        pattern += "green"
        pattern += "grey"
        pattern += "gold"
        pattern += "ivory"
        pattern += "khaki"
        pattern += "oldlace"

        pattern.needle_change(needle=4)
        pattern += (0, 0), (0, 100), (100, 100), (100, 0), (0, 0)
        pattern.add_command(MATRIX_TRANSLATE, 25, 25)
        pattern.add_command(MATRIX_ROTATE, 360.0 / 3)

        pattern.needle_change(needle=5)
        pattern += (0, 0), (0, 100), (100, 100), (100, 0), (0, 0)
        pattern.add_command(MATRIX_TRANSLATE, 25, 25)
        pattern.add_command(MATRIX_ROTATE, 360.0 / 3)

        pattern.needle_change(needle=6)
        pattern += (0, 0), (0, 100), (100, 100), (100, 0), (0, 0)

        write_tbf(pattern, file1, settings={"ct0": file2})
        f_pattern = read_tbf(file1)

        self.assertNotEqual(len(f_pattern.threadlist), 1)
        self.assertEqual(f_pattern.count_stitch_commands(STITCH), 3 * 5)

        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)



import pyembroidery
from pyembroidery import (
    CONTINGENCY_SEQUIN_UTILIZE,
    decode_embroidery_command,
    STITCH,
    COLOR_CHANGE,
    NEEDLE_SET,
    TRIM,
    JUMP,
    SEQUIN_MODE,
    SEQUIN_EJECT,
    STOP,
    SLOW,
    FAST,
    END,
    get_common_name_dictionary,
    COMMAND_MASK,
)
from pyembroidery.WriteHelper import write_string_utf8

WRITES_SPEEDS = True
SEQUIN_CONTINGENCY = CONTINGENCY_SEQUIN_UTILIZE


def write(pattern, f, settings=None):
    writer = GenericWriter(pattern, f, settings)
    writer.write()


class GenericWriter:
    """
    Generic Writer will write generic data fit to a formatted set of strings.

    Blocks are established by the first segment.
    Colors are established by the first segment or the first segment after a color change or by a needle_set.
    Documents are established by the first segment, or the first segment after an end.

    Segment is the default for any-command. Specific commands override segment.

    stitch is the default stitch.
    stitch_first overrides stitch for the first stitch in a block.
    stitch_last overrides stitch for the last stitch in a block.

    trims and jumps occurring before a color change belong to the previous color.
    trims and jumps occurring after a color change belong to the next color.

    Missing segments are treated as if they never existed. Value properties will differ if segments are excluded.
    """

    def __init__(self, pattern, f, settings):
        self.pattern = pattern
        self.f = f
        self.settings = settings
        self.metadata_entry = settings.get("metadata_entry", None)
        self.thread_entry = settings.get("thread_entry", None)
        self.pattern_start = settings.get("pattern_start", None)
        self.pattern_end = settings.get("pattern_end", None)
        self.document_start = settings.get("document_start", None)
        self.document_end = settings.get("document_end", None)
        self.color_start = settings.get("color_start", None)
        self.color_end = settings.get("color_end", None)
        self.color_join = settings.get("color_join", None)
        self.block_start = settings.get("block_start", None)
        self.block_end = settings.get("block_end", None)
        self.block_join = settings.get("block_join", None)
        self.segment_start = settings.get("segment_start", None)
        self.segment = settings.get("segment", None)
        self.segment_end = settings.get("segment_end", None)
        self.segment_join = settings.get("segment_join", None)
        self.stitch_first = settings.get("stitch_first", None)
        self.stitch_last = settings.get("stitch_last", None)
        self.stitch = settings.get("stitch", None)
        self.stop = settings.get("stop", None)
        self.jump = settings.get("jump", None)
        self.trim = settings.get("trim", None)
        self.needle_set = settings.get("needle_set", None)
        self.color_change = settings.get("color_change", None)
        self.sequin = settings.get("sequin", None)
        self.sequin_mode = settings.get("sequin_mode", None)
        self.slow = settings.get("slow", None)
        self.fast = settings.get("fast", None)
        self.end = settings.get("end", None)

        self.format_dictionary = {}
        self.pattern_established = False
        self.document_established = False
        self.color_established = False
        self.block_established = False
        self.document_index = -1
        self.thread = None
        self.thread_index = -1
        self.stitch_index = -1
        self.color_index = -1
        self.block_index = -1
        self.dx = 0
        self.dy = 0
        self.xx = 0
        self.yy = 0
        self.last_x = 0
        self.last_y = 0
        self.z = 0.0
        self.z_increment = settings.get("stitch_z_travel", 10.0)
        self.command_index = 0

        self.current_stitch = None
        self.x = None
        self.y = None
        self.command = None
        self.cmd = None
        self.thread = None
        self.needle = None
        self.order = None
        self.cmd_str = None

        self.block_closing = False
        self.color_closing = False
        self.document_closing = False

        self.block_opening = False
        self.color_opening = False
        self.document_opening = False

    def write_opens(self):
        if self.document_opening:
            self.document_opening = False
            if self.document_start is not None:
                write_string_utf8(self.f, self.document_start.format_map(self.format_dictionary))
        if self.color_opening:
            self.color_opening = False
            if self.color_join is not None and self.color_index != 0:
                write_string_utf8(self.f, self.color_join.format_map(self.format_dictionary))
            if self.color_start is not None:
                write_string_utf8(self.f, self.color_start.format_map(self.format_dictionary))
        if self.block_opening:
            self.block_opening = False
            if self.block_join is not None and self.block_index != 0:
                write_string_utf8(self.f, self.block_join.format_map(self.format_dictionary))
            if self.block_start is not None:
                write_string_utf8(self.f, self.block_start.format_map(self.format_dictionary))

    def write_closes(self):
        if self.block_closing:
            self.block_closing = False
            if self.block_end is not None:
                write_string_utf8(self.f, self.block_end.format_map(self.format_dictionary))
        if self.color_closing:
            self.color_closing = False
            if self.color_end is not None:
                write_string_utf8(self.f, self.color_end.format_map(self.format_dictionary))
        if self.document_closing:
            self.document_closing = False
            if self.document_end is not None:
                write_string_utf8(self.f, self.document_end.format_map(self.format_dictionary))

    def get_write_segment(self, cmd):
        # SEQUIN_MODE
        if cmd == SEQUIN_MODE and self.sequin_mode is not None:
            return self.sequin_mode

        # SEQUIN
        if cmd == SEQUIN_EJECT and self.sequin is not None:
            return self.sequin

        # STITCH
        if cmd == STITCH and self.stitch is not None:
            return self.stitch

        # TRIM
        if cmd == TRIM and self.trim is not None:
            return self.trim

        # JUMP
        if cmd == JUMP and self.jump is not None:
            return self.jump

        # COLOR CHANGE
        if cmd == COLOR_CHANGE and self.color_change is not None:
            return self.color_change

        # NEEDLE SET
        if cmd == NEEDLE_SET and self.needle_set is not None:
            return self.needle_set

        # STOP COMMAND
        if cmd == STOP and self.stop is not None:
            return self.stop

        # SLOW COMMAND
        if cmd == SLOW and self.slow is not None:
            return self.slow

        # FAST COMMAND
        if cmd == FAST and self.fast is not None:
            return self.fast

        # END COMMAND
        if cmd == END and self.end is not None:
            return self.end

        # GENERIC SEGMENT
        return self.segment

    def set_document_statistics(self):
        pattern = self.pattern
        self.format_dictionary.update(pattern.extras)

        bounds = pattern.bounds()  # convert to mm.
        width = bounds[2] - bounds[0]
        height = bounds[3] - bounds[1]

        stitch_counts = {}
        for s in pattern.stitches:
            command = s[2] & COMMAND_MASK
            if command in stitch_counts:
                stitch_counts[command] += 1
            else:
                stitch_counts[command] = 1

        names = get_common_name_dictionary()
        for name in names:
            value = names[name].lower()
            self.format_dictionary[value + "_count"] = stitch_counts.get(name,0)
        self.format_dictionary.update(
            {
                "stitch_total": pattern.count_stitches(),
                "thread_total": pattern.count_threads(),
                "extents_left": bounds[0],
                "extends_top": bounds[1],
                "extends_right": bounds[2],
                "extends_bottom": bounds[3],
                "extends_width": width,
                "extends_height": height,
                "extents_left_mm": bounds[0] / 10.0,
                "extends_top_mm": bounds[1] / 10.0,
                "extends_right_mm": bounds[2] / 10.0,
                "extends_bottom_mm": bounds[3] / 10.0,
                "extends_width_mm": width / 10.0,
                "extends_height_mm": height / 10.0,
            }
        )

    def update_positions(self, x, y, cmd):
        self.dx = x - self.last_x
        self.dy = y - self.last_y
        idx = int(round(x - self.xx))
        idy = int(round(y - self.yy))
        self.xx += idx
        self.yy += idy
        self.format_dictionary.update(
            {
                "x": x,
                "y": y,
                "z": self.z,
                "_x": -x,
                "_y": -y,
                "dx": self.dx,
                "dy": self.dy,
                "idx": idx,
                "idy": idy,
                "_idx": -idx,
                "_idy": -idy,
                "ix": self.xx,
                "iy": self.yy,
                "_ix": -self.xx,
                "_iy": -self.yy,
                "last_x": self.last_x,
                "last_y": self.last_y,
                "_last_x": -self.last_x,
                "_last_y": -self.last_y,
            }
        )
        if cmd == STITCH:
            self.z += self.z_increment
        self.last_x = x
        self.last_y = y

    def update_command(self):
        try:
            self.current_stitch = self.pattern.stitches[self.command_index]

            self.x, self.y, self.command = self.current_stitch
            self.cmd, self.thread, self.needle, self.order = decode_embroidery_command(
                self.command
            )
            self.cmd_str = pyembroidery.get_common_name_dictionary()[self.cmd]
        except IndexError:
            self.current_stitch = None
            self.x = None
            self.y = None
            self.command = None
            self.cmd = None
            self.thread = None
            self.needle = None
            self.order = None
            self.cmd_str = None
        self.format_dictionary.update(
            {
                "index": self.command_index,
                "command": self.command,
                "cmd_str": self.cmd_str,
                "cmd": self.cmd,
                "cmd_thread": self.thread,
                "cmd_needle": self.needle,
                "cmd_order": self.order,
            }
        )

    def open_pattern(self):
        if not self.pattern_established:
            self.pattern_established = True
            if self.pattern_start is not None:
                write_string_utf8(
                    self.f, self.pattern_start.format_map(self.format_dictionary)
                )

    def open_document(self):
        # DOCUMENT START
        if not self.document_established:
            self.document_established = True
            self.document_index += 1
            self.document_opening = True
            self.color_index = 0

            self.format_dictionary.update(
                {
                    "document_index": self.document_index,
                    "document_index1": self.document_index + 1,
                    "color_index": self.color_index,
                    "color_index1": self.color_index + 1,
                    "block_index": self.block_index,
                    "block_index1": self.block_index + 1,
                }
            )

    def open_color(self):
        # COLOR START
        if not self.color_established:
            self.color_established = True
            self.thread_index += 1
            self.color_opening = True

            self.thread = self.pattern.get_thread_or_filler(self.thread_index)
            self.block_index = 0
            self.color_index += 1
            self.format_dictionary.update(
                {
                    "document_index": self.document_index,
                    "document_index1": self.document_index + 1,
                    "color_index": self.color_index,
                    "color_index1": self.color_index + 1,
                    "block_index": self.block_index,
                    "block_index1": self.block_index + 1,
                }
            )

    def open_block(self):
        # BLOCK START
        if not self.block_established:
            self.block_established = True
            self.block_index += 1
            self.block_opening = True
            self.format_dictionary.update(
                {
                    "document_index": self.document_index,
                    "document_index1": self.document_index + 1,
                    "color_index": self.color_index,
                    "color_index1": self.color_index + 1,
                    "block_index": self.block_index,
                    "block_index1": self.block_index + 1,
                }
            )

    def write_segment(self, segment):
        # SEGMENT
        if self.segment_start is not None:
            write_string_utf8(self.f, self.segment_start.format_map(self.format_dictionary))

        write_string_utf8(self.f, segment.format_map(self.format_dictionary))

        # SEGMENT JOIN
        if self.segment_join is not None:
            write_string_utf8(self.f, self.segment_join.format_map(self.format_dictionary))

        # SEGMENT_END
        if self.segment_end is not None:
            write_string_utf8(self.f, self.segment_end.format_map(self.format_dictionary))

    def close_pattern(self):
        if self.pattern_established:
            self.pattern_established = False
            if self.pattern_end is not None:
                write_string_utf8(
                    self.f, self.pattern_end.format_map(self.format_dictionary)
                )

    def close_document(self):
        # DOCUMENT END
        if self.document_established:
            self.document_established = False
            self.document_closing = True

    def close_color(self):
        # COLOR END
        if self.color_established:
            self.color_established = False
            self.color_closing = True

    def close_block(self):
        # BLOCK END
        if self.block_established:
            self.block_established = False
            self.block_closing = True

    def write(self):
        # DOCUMENT STATISTICS
        self.set_document_statistics()

        self.open_pattern()
        if self.metadata_entry is not None:
            for i, key in enumerate(self.pattern.extras):
                value = self.pattern.extras[key]
                self.format_dictionary.update({
                    "metadata_index": i,
                    "metadata_key": str(key),
                    "metadata_value": str(value),
                })
                write_string_utf8(
                    self.f, self.metadata_entry.format_map(self.format_dictionary)
                )

        if self.thread_entry is not None:
            for i, thread in enumerate(self.pattern.threadlist):
                self.format_dictionary.update({
                    "thread_index": i,
                    "thread_color": thread.hex_color(),
                    "thread_description": thread.description,
                    "thread_brand": thread.brand,
                    "thread_catalog_number": thread.catalog_number,
                    "thread_chart": thread.chart,
                    "thread_details": thread.details,
                    "thread_weight": thread.weight,
                    "thread_red": thread.get_red(),
                    "thread_green": thread.get_green(),
                    "thread_blue": thread.get_blue(),
                })
                write_string_utf8(
                    self.f, self.thread_entry.format_map(self.format_dictionary)
                )
        for self.command_index in range(0, len(self.pattern.stitches)):
            self.update_command()
            write_segment = self.get_write_segment(self.cmd)

            # MAIN CODE, there is something to write.
            if write_segment is not None:
                if isinstance(write_segment, dict):
                    key, default = write_segment[None]
                    key = key.format_map(self.format_dictionary)
                    write_segment = write_segment.get(key, default)
                self.update_positions(self.x, self.y, self.cmd)
                if self.cmd == SEQUIN_MODE:
                    self.open_document()
                    self.open_color()
                    self.open_block()

                # SEQUIN
                if self.cmd == SEQUIN_EJECT:
                    self.open_document()
                    self.open_color()
                    self.open_block()

                # STITCH
                if self.cmd == STITCH:
                    self.open_document()
                    self.open_color()
                    self.open_block()

                # TRIM
                if self.cmd == TRIM:
                    self.open_document()
                    self.open_color()

                # JUMP
                if self.cmd == JUMP:
                    self.open_document()
                    self.open_color()
                    self.open_block()

                # COLOR CHANGE
                if self.cmd == COLOR_CHANGE:
                    self.open_document()

                # NEEDLE SET
                if self.cmd == NEEDLE_SET:
                    self.open_document()
                    self.open_color()

                # STOP COMMAND
                if self.cmd == STOP:
                    self.open_document()
                    self.open_color()
                    self.open_block()

                # SLOW COMMAND
                if self.cmd == SLOW:
                    self.open_document()
                    self.open_color()
                    self.open_block()

                # FAST COMMAND
                if self.cmd == FAST:
                    self.open_document()
                    self.open_color()
                    self.open_block()

                # END COMMAND
                if self.cmd == END:
                    self.open_document()
                    self.open_color()
                    self.open_block()

                self.write_opens()
                self.write_segment(write_segment)

                if self.cmd == SEQUIN_MODE:
                    pass

                # SEQUIN
                if self.cmd == SEQUIN_EJECT:
                    pass

                # STITCH
                if self.cmd == STITCH:
                    pass

                # TRIM
                if self.cmd == TRIM:
                    self.close_block()

                # JUMP
                if self.cmd == JUMP:
                    pass

                # COLOR CHANGE
                if self.cmd == COLOR_CHANGE:
                    self.close_block()
                    self.close_color()

                # NEEDLE SET
                if self.cmd == NEEDLE_SET:
                    pass

                # STOP COMMAND
                if self.cmd == STOP:
                    pass

                # SLOW COMMAND
                if self.cmd == SLOW:
                    pass

                # FAST COMMAND
                if self.cmd == FAST:
                    pass

                # END COMMAND
                if self.cmd == END:
                    self.close_block()
                    self.close_color()
                    self.close_document()
                self.write_closes()

        self.close_block()
        self.close_color()
        self.close_document()
        self.close_pattern()



from . import decode_embroidery_command
from .EmbConstant import *
from .WriteHelper import write_string_utf8, write_int_8

FULL_JUMP = False
ROUND = True
MAX_JUMP_DISTANCE = 127
MAX_STITCH_DISTANCE = 127
THREAD_CHANGE_COMMAND = NEEDLE_SET
EXPLICIT_TRIM = True


def write(pattern, f, settings=None):
    if settings is not None and "ct0" in settings:
        ct0 = settings.get("ct0")
        write_ct0(pattern, ct0)
    bounds = pattern.bounds()

    name = pattern.get_metadata("name", "Untitled")
    write_string_utf8(f, "3.00")
    for i in range(f.tell(), 0x80):
        f.write(b"\x20")  # space
    write_string_utf8(f, "LA:%-16s\r" % name)
    write_string_utf8(f, "ST:%7d\r" % pattern.count_stitches())
    write_string_utf8(f, "CO:%3d\r" % pattern.count_needle_sets())

    write_string_utf8(f, "+X:%5d\r" % abs(bounds[2]))
    write_string_utf8(f, "-X:%5d\r" % abs(bounds[0]))
    write_string_utf8(f, "+Y:%5d\r" % abs(bounds[3]))
    write_string_utf8(f, "-Y:%5d\r" % abs(bounds[1]))
    ax = 0
    ay = 0
    if len(pattern.stitches) > 0:
        last = len(pattern.stitches) - 1
        ax = int(pattern.stitches[last][0])
        ay = -int(pattern.stitches[last][1])
    if ax >= 0:
        write_string_utf8(f, "AX:+%5d\r" % ax)
    else:
        write_string_utf8(f, "AX:-%5d\r" % abs(ax))
    if ay >= 0:
        write_string_utf8(f, "AY:+%5d\r" % ay)
    else:
        write_string_utf8(f, "AY:-%5d\r" % abs(ay))

    # TP is unknown.
    tp = pattern.get_metadata("tp", "EG/")
    write_string_utf8(f, "TP:%-32s\r" % tp)

    # JC is unknown.
    jc = "3"
    write_string_utf8(f, "JC:%s\r" % jc)

    # DO is the thread order.
    write_string_utf8(f, "DO:")
    thread_order = [0] * 0x100
    index = 0
    for stitch in pattern.stitches:
        data = stitch[2] & COMMAND_MASK
        if data == NEEDLE_SET:
            flag, thread, needle, order = decode_embroidery_command(stitch[2])
            thread_order[index] = needle
            index += 1
    for n in thread_order:
        write_int_8(f, n)
    write_string_utf8(f, "\r")

    # DA is the threadlist. This is not *only* the used threads but any threads in the set.
    write_string_utf8(f, "DA:")
    if len(pattern.threadlist) > 0:
        for thread in pattern.threadlist:
            write_int_8(f, 0x45)
            write_int_8(f, thread.get_red())
            write_int_8(f, thread.get_green())
            write_int_8(f, thread.get_blue())
            write_int_8(f, 0x20)

    # Padding to 501
    for i in range(f.tell(), 0x376):
        f.write(b"\x20")  # space

    # Seen in only some files.
    f.write(b"\x0d\x1A")

    # Pad to the end of the header.
    for i in range(f.tell(), 0x600):
        f.write(b"\x20")  # space
    # END HEADER

    stitches = pattern.stitches
    xx = 0
    yy = 0
    for stitch in stitches:
        x = stitch[0]
        y = stitch[1]
        data = stitch[2] & COMMAND_MASK
        dx = int(round(x - xx))
        dy = int(round(y - yy))
        xx += dx
        yy += dy

        if data == STITCH:
            cmd = 0x80
            f.write(bytes(bytearray([dx & 0xFF, -dy & 0xFF, cmd])))
        elif data == JUMP:
            cmd = 0x90
            f.write(bytes(bytearray([dx & 0xFF, -dy & 0xFF, cmd])))
        elif data == STOP:
            cmd = 0x40
            f.write(bytes(bytearray([dx & 0xFF, -dy & 0xFF, cmd])))
        elif data == TRIM:
            cmd = 0x86
            f.write(bytes(bytearray([dx & 0xFF, -dy & 0xFF, cmd])))
        elif data == NEEDLE_SET:
            cmd = 0x81
            f.write(bytes(bytearray([dx & 0xFF, -dy & 0xFF, cmd])))
        elif data == END:
            cmd = 0x8F
            f.write(bytes(bytearray([dx & 0xFF, -dy & 0xFF, cmd])))
            break
    # Terminal character.
    f.write(b"\x1a")


def write_ct0(pattern, filename, settings=None):
    with open(filename, "wb") as f:
        _write_ct0(pattern, f, settings=settings)


def _write_ct0(pattern, f, settings=None):
    write_string_utf8(f, "TAJ-DGML-PULSE  1-1A 2060(550.0")
    write_int_8(f, 0x81)
    write_string_utf8(f, "~400.0)S         2.00")
    for i in range(f.tell(), 0x60):
        f.write(b"\x20")
    write_string_utf8(f, "DC1:100\rDC2:100\rDC3:  0\rDC4:N\rDC5:S\r")
    for i in range(f.tell(), 0x108):
        f.write(b"\x20")
    write_string_utf8(f, "NS1:11")
    index = 0
    for stitch in pattern.stitches:
        data = stitch[2] & COMMAND_MASK
        if data == NEEDLE_SET:
            flag, thread, needle, order = decode_embroidery_command(stitch[2])
            write_int_8(f, needle + 0x30)
            write_int_8(f, 0x31)
            index += 1
    for i in range(f.tell(), 0x30D):
        f.write(b"\x20")
    write_string_utf8(f, "\rRP0:N\rRP1:  \rRP2:  \rRP3:      \rRP4:      \rRP5: \rRP6: \rRP7: \rST1:")
    for i in range(f.tell(), 0x434):
        f.write(b"\x20")
    write_string_utf8(f, "ST0:0\rAO1:0\rAO2:0\rAO3:0\rOF1:            \rOF2:            \rOF3:            \rNS2:")
    for i in range(index):
        write_int_8(f, 0x30)
    for i in range(f.tell(), 0x583):
        f.write(b"\x20")
    write_string_utf8(f, "\rNS3:")
    for i in range(f.tell(), 0x778):
        f.write(b"\x20")
    write_string_utf8(f, "\r\x1A")
    for i in range(f.tell(), 0x790):
        f.write(b"\x20")



from __future__ import print_function

import unittest
from test.pattern_for_tests import *


class TestDataCatalog(unittest.TestCase):

    def test_catalog_files(self):
        for f in EmbPattern.supported_formats():
            self.assertIn("extensions", f)
            self.assertIn("extension", f)
            self.assertIn("description", f)
            self.assertIn("category", f)



import unittest

from test.pattern_for_tests import *


class TestExplicitIOErrors(unittest.TestCase):
    def test_read_non_file(self):
        """
        1.5.0 adds explicit error raising.
        We test that now.
        """
        file1 = "nosuchfile.dst"
        self.assertRaises(IOError, lambda: EmbPattern(file1))

    def test_write_non_supported(self):
        """
        1.5.0 adds explicit error raising.
        We test that now.
        """
        pattern = get_simple_pattern()
        file1 = "nosuchfile.pdf"
        self.assertRaises(IOError, lambda: pattern.write(file1))

    def test_write_no_writer(self):
        """
        1.5.0 adds explicit error raising.
        We test that now.
        """
        pattern = get_simple_pattern()
        file1 = "nosuchfile.dat"
        self.assertRaises(IOError, lambda: pattern.write(file1))



from __future__ import print_function

import unittest

from test.pattern_for_tests import *


class TestConverts(unittest.TestCase):

    def position_equals(self, stitches, j, k):
        self.assertEqual(stitches[j][:1], stitches[k][:1])

    def test_convert_pec_to_u01(self):
        file1 = "convert_u01.pec"
        file2 = "converted_pec.u01"
        write_pec(get_big_pattern(), file1)
        f_pattern = read_pec(file1)
        write_u01(f_pattern, file2)
        t_pattern = read_u01(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(NEEDLE_SET), 16)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("pec->u01: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_convert_pec_to_csv(self):
        file1 = "convert_csv.pec"
        file2 = "converted_pec.csv"
        write_pec(get_big_pattern(), file1)
        f_pattern = read_pec(file1)
        write_csv(f_pattern, file2)
        t_pattern = read_csv(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("pec->csv: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_convert_pec_to_exp(self):
        file1 = "convert_exp.pec"
        file2 = "converted_pec.exp"
        write_pec(get_big_pattern(), file1)
        f_pattern = read_pec(file1)
        write_exp(f_pattern, file2)
        t_pattern = read_exp(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("pec->exp: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_convert_pec_to_pes(self):
        file1 = "convert_pes.pec"
        file2 = "converted_pec.pes"
        write_pec(get_big_pattern(), file1)
        f_pattern = read_pec(file1)
        write_pes(f_pattern, file2)
        t_pattern = read_pes(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("pec->pes: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_convert_pec_to_jef(self):
        file1 = "convert_jef.pec"
        file2 = "converted_pec.jef"
        write_pec(get_big_pattern(), file1)
        f_pattern = read_pec(file1)
        write_jef(f_pattern, file2)
        t_pattern = read_jef(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("pec->jef: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_convert_pec_to_pec(self):
        file1 = "convert_pec.pec"
        file2 = "converted_pec.pec"
        write_pec(get_big_pattern(), file1)
        f_pattern = read_pec(file1)
        write_pec(f_pattern, file2)
        t_pattern = read_pec(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("pec->pec: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_convert_pec_to_vp3(self):
        file1 = "convert_vp3.pec"
        file2 = "converted_pec.vp3"
        write_pec(get_big_pattern(), file1)
        f_pattern = read_pec(file1)
        write_vp3(f_pattern, file2)
        t_pattern = read_vp3(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("pec->vp3: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_convert_pec_to_dst(self):
        file1 = "convert_dst.pec"
        file2 = "converted_pec.dst"
        write_pec(get_big_pattern(), file1)
        f_pattern = read_pec(file1)
        write_dst(f_pattern, file2)
        t_pattern = read_dst(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("pec->dst: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_convert_pec_to_gcode(self):
        file1 = "convert_gcode.pec"
        file2 = "converted_pec.gcode"
        write_pec(get_big_pattern(), file1)
        f_pattern = read_pec(file1)
        write_gcode(f_pattern, file2)
        t_pattern = read_gcode(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("pec->gcode: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)
        
    def test_convert_pec_to_xxx(self):
        file1 = "convert_xxx.pec"
        file2 = "converted_pec.xxx"
        write_pec(get_big_pattern(), file1)
        f_pattern = read_pec(file1)
        write_xxx(f_pattern, file2)
        t_pattern = read_xxx(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("pec->xxx: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_convert_pec_to_tbf(self):
        file1 = "convert_tbf.pec"
        file2 = "converted_pec.tbf"
        write_pec(get_big_pattern(), file1)
        f_pattern = read_pec(file1)
        write_tbf(f_pattern, file2)
        t_pattern = read_tbf(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(NEEDLE_SET), 16)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("pec->tbf: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)


import struct
import zlib
from math import sqrt

from .EmbConstant import *
from .EmbThread import EmbThread

SEQUIN_CONTINGENCY = CONTINGENCY_SEQUIN_STITCH
FULL_JUMP = True

# Static characters for writing to image.
characters = {
    "0": [
        [9, 9, 4, 1, 0, 1, 5, 9, 9],
        [9, 3, 0, 0, 0, 0, 0, 5, 9],
        [7, 0, 1, 7, 9, 6, 0, 0, 9],
        [4, 0, 5, 9, 9, 9, 4, 0, 6],
        [3, 0, 8, 9, 9, 9, 6, 0, 4],
        [2, 0, 9, 5, 0, 6, 8, 0, 3],
        [2, 0, 9, 4, 0, 6, 8, 0, 3],
        [2, 0, 9, 9, 9, 9, 8, 0, 3],
        [3, 0, 8, 9, 9, 9, 6, 0, 4],
        [4, 0, 5, 9, 9, 9, 4, 0, 6],
        [7, 0, 1, 7, 9, 6, 0, 0, 8],
        [9, 3, 0, 0, 0, 0, 0, 5, 9],
        [9, 9, 4, 1, 0, 1, 5, 9, 9],
    ],
    "1": [
        [9, 8, 5, 2, 0, 1, 9, 9, 9],
        [9, 1, 0, 0, 0, 1, 9, 9, 9],
        [9, 2, 4, 7, 1, 1, 9, 9, 9],
        [9, 9, 9, 9, 1, 1, 9, 9, 9],
        [9, 9, 9, 9, 1, 1, 9, 9, 9],
        [9, 9, 9, 9, 1, 1, 9, 9, 9],
        [9, 9, 9, 9, 1, 1, 9, 9, 9],
        [9, 9, 9, 9, 1, 1, 9, 9, 9],
        [9, 9, 9, 9, 1, 1, 9, 9, 9],
        [9, 9, 9, 9, 1, 1, 9, 9, 9],
        [9, 9, 9, 9, 1, 1, 9, 9, 9],
        [9, 3, 0, 0, 0, 0, 0, 0, 3],
        [9, 3, 0, 0, 0, 0, 0, 0, 3],
    ],
    "2": [
        [8, 5, 2, 1, 1, 2, 6, 9, 9],
        [4, 0, 0, 0, 0, 0, 0, 4, 9],
        [5, 4, 7, 8, 8, 5, 0, 0, 8],
        [9, 9, 9, 9, 9, 9, 3, 0, 7],
        [9, 9, 9, 9, 9, 9, 3, 0, 9],
        [9, 9, 9, 9, 9, 8, 0, 4, 9],
        [9, 9, 9, 9, 8, 1, 2, 9, 9],
        [9, 9, 9, 8, 2, 2, 9, 9, 9],
        [9, 9, 8, 1, 2, 9, 9, 9, 9],
        [9, 8, 1, 2, 9, 9, 9, 9, 9],
        [7, 1, 2, 9, 9, 9, 9, 9, 9],
        [3, 0, 0, 0, 0, 0, 0, 0, 6],
        [3, 0, 0, 0, 0, 0, 0, 0, 6],
    ],
    "3": [
        [9, 6, 2, 1, 1, 2, 6, 9, 9],
        [5, 0, 0, 0, 0, 0, 0, 4, 9],
        [6, 3, 7, 8, 9, 6, 1, 0, 8],
        [9, 9, 9, 9, 9, 9, 3, 0, 8],
        [9, 9, 9, 9, 8, 6, 1, 1, 9],
        [9, 9, 4, 0, 0, 0, 2, 8, 9],
        [9, 9, 4, 0, 0, 0, 2, 8, 9],
        [9, 9, 9, 9, 8, 6, 1, 1, 8],
        [9, 9, 9, 9, 9, 9, 5, 0, 6],
        [9, 9, 9, 9, 9, 9, 5, 0, 5],
        [3, 5, 7, 9, 8, 6, 1, 0, 7],
        [2, 0, 0, 0, 0, 0, 0, 3, 9],
        [8, 4, 2, 0, 1, 2, 5, 9, 9],
    ],
    "4": [
        [9, 9, 9, 9, 7, 0, 0, 7, 9],
        [9, 9, 9, 9, 2, 1, 0, 7, 9],
        [9, 9, 9, 5, 2, 4, 0, 7, 9],
        [9, 9, 8, 1, 7, 4, 0, 7, 9],
        [9, 9, 3, 4, 9, 4, 0, 7, 9],
        [9, 6, 1, 9, 9, 4, 0, 7, 9],
        [8, 1, 6, 9, 9, 4, 0, 7, 9],
        [4, 3, 9, 9, 9, 4, 0, 7, 9],
        [0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 0],
        [9, 9, 9, 9, 9, 4, 0, 7, 9],
        [9, 9, 9, 9, 9, 4, 0, 7, 9],
        [9, 9, 9, 9, 9, 4, 0, 7, 9],
    ],
    "5": [
        [7, 0, 0, 0, 0, 0, 0, 5, 9],
        [7, 0, 0, 0, 0, 0, 0, 5, 9],
        [7, 0, 5, 9, 9, 9, 9, 9, 9],
        [7, 0, 5, 9, 9, 9, 9, 9, 9],
        [7, 0, 1, 1, 0, 2, 6, 9, 9],
        [7, 0, 0, 0, 0, 0, 0, 5, 9],
        [7, 4, 7, 9, 8, 4, 0, 0, 9],
        [9, 9, 9, 9, 9, 9, 3, 0, 6],
        [9, 9, 9, 9, 9, 9, 5, 0, 6],
        [9, 9, 9, 9, 9, 9, 3, 0, 6],
        [3, 5, 8, 9, 8, 5, 0, 0, 9],
        [2, 0, 0, 0, 0, 0, 0, 6, 9],
        [8, 4, 1, 0, 1, 2, 7, 9, 9],
    ],
    "6": [
        [9, 9, 7, 2, 1, 0, 3, 8, 9],
        [9, 6, 0, 0, 0, 0, 0, 3, 9],
        [8, 0, 0, 4, 8, 9, 6, 4, 9],
        [5, 0, 4, 9, 9, 9, 9, 9, 9],
        [3, 0, 8, 9, 9, 9, 9, 9, 9],
        [2, 1, 8, 3, 0, 1, 3, 8, 9],
        [2, 2, 2, 0, 0, 0, 0, 1, 8],
        [2, 0, 2, 7, 9, 8, 2, 0, 5],
        [3, 0, 7, 9, 9, 9, 7, 0, 3],
        [4, 0, 7, 9, 9, 9, 7, 0, 3],
        [7, 0, 2, 7, 9, 8, 2, 0, 5],
        [9, 3, 0, 0, 0, 0, 0, 1, 8],
        [9, 9, 4, 1, 0, 1, 3, 8, 9],
    ],
    "7": [
        [2, 0, 0, 0, 0, 0, 0, 0, 5],
        [2, 0, 0, 0, 0, 0, 0, 0, 7],
        [9, 9, 9, 9, 9, 9, 3, 2, 9],
        [9, 9, 9, 9, 9, 8, 0, 5, 9],
        [9, 9, 9, 9, 9, 4, 0, 8, 9],
        [9, 9, 9, 9, 9, 1, 3, 9, 9],
        [9, 9, 9, 9, 6, 0, 6, 9, 9],
        [9, 9, 9, 9, 2, 1, 9, 9, 9],
        [9, 9, 9, 7, 0, 4, 9, 9, 9],
        [9, 9, 9, 3, 0, 8, 9, 9, 9],
        [9, 9, 8, 0, 2, 9, 9, 9, 9],
        [9, 9, 5, 0, 6, 9, 9, 9, 9],
        [9, 9, 1, 0, 9, 9, 9, 9, 9],
    ],
    "8": [
        [9, 8, 3, 1, 0, 1, 4, 8, 9],
        [8, 1, 0, 0, 0, 0, 0, 1, 9],
        [5, 0, 3, 8, 9, 7, 2, 0, 6],
        [5, 0, 7, 9, 9, 9, 5, 0, 6],
        [7, 0, 3, 8, 9, 7, 2, 1, 8],
        [9, 7, 2, 0, 0, 0, 2, 7, 9],
        [9, 5, 1, 0, 0, 0, 1, 6, 9],
        [6, 0, 3, 8, 9, 7, 2, 0, 7],
        [2, 0, 9, 9, 9, 9, 7, 0, 4],
        [2, 0, 9, 9, 9, 9, 7, 0, 3],
        [3, 0, 3, 8, 9, 7, 2, 0, 4],
        [7, 0, 0, 0, 0, 0, 0, 1, 8],
        [9, 8, 3, 1, 0, 1, 4, 8, 9],
    ],
    "9": [
        [9, 7, 3, 0, 0, 2, 6, 9, 9],
        [7, 0, 0, 0, 0, 0, 0, 5, 9],
        [3, 0, 4, 8, 9, 7, 1, 0, 9],
        [1, 0, 9, 9, 9, 9, 5, 0, 6],
        [1, 1, 9, 9, 9, 9, 5, 0, 5],
        [3, 0, 4, 8, 9, 7, 1, 0, 4],
        [7, 0, 0, 0, 0, 0, 3, 0, 4],
        [9, 7, 2, 0, 1, 4, 7, 0, 4],
        [9, 9, 9, 9, 9, 9, 5, 0, 5],
        [9, 9, 9, 9, 9, 9, 2, 0, 7],
        [9, 3, 7, 9, 8, 3, 0, 2, 9],
        [9, 1, 0, 0, 0, 0, 0, 7, 9],
        [9, 7, 2, 0, 1, 3, 8, 9, 9],
    ],
    "-": [
        [9, 9, 9, 9, 9, 9, 9, 9, 9],
        [9, 9, 9, 9, 9, 9, 9, 9, 9],
        [9, 9, 9, 9, 9, 9, 9, 9, 9],
        [9, 9, 9, 9, 9, 9, 9, 9, 9],
        [9, 9, 9, 9, 9, 9, 9, 9, 9],
        [9, 9, 9, 9, 9, 9, 9, 9, 9],
        [9, 9, 9, 9, 9, 9, 9, 9, 9],
        [9, 9, 1, 0, 0, 0, 3, 9, 9],
        [9, 9, 1, 0, 0, 0, 3, 9, 9],
        [9, 9, 9, 9, 9, 9, 9, 9, 9],
        [9, 9, 9, 9, 9, 9, 9, 9, 9],
        [9, 9, 9, 9, 9, 9, 9, 9, 9],
        [9, 9, 9, 9, 9, 9, 9, 9, 9],
    ],
    "m": [
        [9, 9, 9, 9, 9, 9, 9, 9, 9],
        [9, 9, 9, 9, 9, 9, 9, 9, 9],
        [9, 9, 9, 9, 9, 9, 9, 9, 9],
        [0, 5, 2, 1, 7, 6, 1, 2, 8],
        [0, 1, 0, 0, 1, 1, 0, 0, 3],
        [0, 2, 9, 4, 0, 4, 8, 2, 1],
        [0, 5, 9, 6, 0, 7, 9, 4, 1],
        [0, 5, 9, 6, 0, 7, 9, 4, 0],
        [0, 5, 9, 6, 0, 7, 9, 4, 0],
        [0, 5, 9, 6, 0, 7, 9, 4, 0],
        [0, 5, 9, 6, 0, 7, 9, 4, 0],
        [0, 5, 9, 6, 0, 7, 9, 4, 0],
        [0, 5, 9, 6, 0, 7, 9, 4, 0],
    ],
}


def write_png(buf, width, height):
    """
    Writes PNG file to disk. Buffer must be RGBA * width * height
    """
    width_byte_4 = width * 4
    raw_data = b"".join(
        b"\x00" + buf[span : span + width_byte_4]
        for span in range(0, height * width * 4, width_byte_4)
    )

    def png_pack(png_tag, data):
        chunk_head = png_tag + data
        return (
            struct.pack("!I", len(data))
            + chunk_head
            + struct.pack("!I", 0xFFFFFFFF & zlib.crc32(chunk_head))
        )

    return b"".join(
        [
            b"\x89PNG\r\n\x1a\n",
            png_pack(b"IHDR", struct.pack("!2I5B", width, height, 8, 6, 0, 0, 0)),
            png_pack(b"IDAT", zlib.compress(raw_data, 9)),
            png_pack(b"IEND", b""),
        ]
    )


class PngBuffer:
    def __init__(self, width, height):
        self.width = int(width + 3)
        self.height = int(height + 3)
        self.buf = bytearray(4 * self.width * self.height)
        self.line_width = 3
        self.fancy = True
        self._red = 0
        self._green = 0
        self._blue = 0
        self._alpha = 0
        self._distance_from_black = 0
        self._gradient_shade_ends = 0.65
        self._gradient_shade_edge = 1.1
        self._gradient_shade_center = 1.55
        self._gradient_color_position1 = 0.40
        self._gradient_color_position2 = 0.50
        self._gradient_color_position3 = 0.70

    def modify_gradient(
        self,
        gradient_shade_ends=0.65,
        gradient_shade_edge=1.1,
        gradient_shade_center=1.55,
        gradient_color_position1=0.40,
        gradient_color_position2=0.50,
        gradient_color_position3=0.70,
    ):
        self._gradient_shade_ends = gradient_shade_ends
        self._gradient_shade_edge = gradient_shade_edge
        self._gradient_shade_center = gradient_shade_center
        self._gradient_color_position1 = gradient_color_position1
        self._gradient_color_position2 = gradient_color_position2
        self._gradient_color_position3 = gradient_color_position3

    def set_color(self, r, g, b, a=255):
        self._red = r
        self._green = g
        self._blue = b
        self._alpha = a
        rmean = int(r / 2)
        self._distance_from_black = sqrt(
            (((512 + rmean) * r * r) >> 8) + 4 * g * g + (((767 - rmean) * b * b) >> 8)
        )

    def gradient(self, position_in_line):
        """
        This function gives different value scales between 0 and 1 which are used to multiply
        all the components of the color to create a darker value. This generally works save for
        black which is already (0,0,0) and can't increase value.
        """
        if (
            position_in_line <= self._gradient_color_position1
        ):  # start_of_grdient -> Position 1
            from_shade = self._gradient_shade_ends
            to_shade = self._gradient_shade_edge
            range = self._gradient_color_position1 - 0  # Range of transition
            amount = (position_in_line - 0) * (1 / range)

        # light to extra_light
        elif (
            position_in_line <= self._gradient_color_position2
        ):  # Position 1 -> Position 2.
            from_shade = self._gradient_shade_edge
            to_shade = self._gradient_shade_center
            range = self._gradient_color_position2 - self._gradient_color_position1
            amount = (position_in_line - self._gradient_color_position1) * (1 / range)

        # center of gradient

        # extra_light to light
        elif (
            position_in_line <= self._gradient_color_position3
        ):  # Position 2 -> Position 3
            from_shade = self._gradient_shade_center
            to_shade = self._gradient_shade_edge
            range = self._gradient_color_position3 - self._gradient_color_position2
            amount = (position_in_line - self._gradient_color_position2) * (1 / range)

        # light to dark
        elif position_in_line <= 1:  # Position 3 -> end_of_gradient
            from_shade = self._gradient_shade_edge
            to_shade = self._gradient_shade_ends
            range = 1 - self._gradient_color_position3
            amount = (position_in_line - self._gradient_color_position3) * (1 / range)
        else:
            raise ValueError("Did not occur within line.")
        v = amount * (to_shade - from_shade) + from_shade
        return max(min(v, 1.0), 0.0)

    def background(self, red, green, blue, alpha):
        for i in range(0, len(self.buf), 4):
            self.buf[i] = red
            self.buf[i + 1] = green
            self.buf[i + 2] = blue
            self.buf[i + 3] = alpha

    def plot(self, x, y, v=None, a=None):
        """
        Plot a particular point in the canvas.
        :param x: x position to plot
        :param y: y position to plot
        :param v: value scale of particular color (0-1)
        :param a: alpha to set for this particular plot if not self.alpha
        :return:
        """
        a = int(a) if a is not None else self._alpha
        if v is None:
            v = 1.0
        try:
            x += 1
            y += 1
            pos = (self.width * y) + x
            idx = pos * 4
            background_a = self.buf[idx + 3]

            # get rgb and check if close to black and make off dark gray
            # this makes black have highlights
            if self._distance_from_black < 15:
                r = 35
                g = 35
                b = 35
                r = r * v
                g = g * v
                b = b * v
                # end of check and make gray
            else:
                r = self._red * v
                g = self._green * v
                b = self._blue * v
            if 0 > r:
                r = 0
            if 0 > g:
                g = 0
            if 0 > b:
                b = 0
            if r > 255:
                r = 255
            if g > 255:
                g = 255
            if b > 255:
                b = 255
            if background_a != 0 and a != 255:
                s_alpha = a / 255.0
                s_background_a = background_a / 255.0
                one_minus_salpha = 1 - s_alpha

                background_r = self.buf[idx]
                background_g = self.buf[idx + 1]
                background_b = self.buf[idx + 2]
                r = r * s_alpha + one_minus_salpha * background_r
                g = g * s_alpha + one_minus_salpha * background_g
                b = b * s_alpha + one_minus_salpha * background_b
                a = (s_alpha + one_minus_salpha * s_background_a) * 255
            self.buf[idx] = int(r)
            self.buf[idx + 1] = int(g)
            self.buf[idx + 2] = int(b)
            self.buf[idx + 3] = (
                int(a) - 4
            )  # remove just a little opacity for background color tint show thru
        except IndexError:
            pass

    def draw_line(self, x0, y0, x1, y1):
        dy = y1 - y0  # BRESENHAM LINE DRAW ALGORITHM
        dx = x1 - x0
        if dy < 0:
            dy = -dy
            step_y = -1
        else:
            step_y = 1
        odx = abs(dx)
        ody = abs(dy)
        if dx < 0:
            dx = -dx
            step_x = -1
        else:
            step_x = 1
        i = 0
        if dx > dy:
            dy <<= 1  # dy is now 2*dy
            dx <<= 1
            fraction = dy - (dx >> 1)  # same as 2*dy - dx
            self.line_for_point(x0, y0, False, odx, i)
            i += 1

            while x0 != x1:
                if fraction >= 0:
                    y0 += step_y
                    fraction -= dx  # same as fraction -= 2*dx
                x0 += step_x
                fraction += dy  # same as fraction += 2*dy
                self.line_for_point(x0, y0, False, odx, i)
                i += 1
        else:
            dy <<= 1  # dy is now 2*dy
            dx <<= 1  # dx is now 2*dx
            fraction = dx - (dy >> 1)
            self.line_for_point(x0, y0, True, ody, i)
            i += 1
            while y0 != y1:
                if fraction >= 0:
                    x0 += step_x
                    fraction -= dy
                y0 += step_y
                fraction += dx
                self.line_for_point(x0, y0, True, ody, i)

    def line_for_point(self, x, y, dy, max_pos, index):
        w = self.line_width
        left = w >> 1
        right = w - left
        if self.fancy and max_pos > 0:
            v = self.gradient(index / max_pos)
        else:
            v = 1.0
        if dy:
            for pos in range(-left, right):
                self.plot(x + pos, y, v)
        else:
            for pos in range(-left, right):
                self.plot(x, y + pos, v)

    def draw_text(self, x, y, string, rotate=False):
        for c in string:
            m = characters[c]
            for cx in range(len(m[0])):
                for cy in range(len(m)):
                    v = m[cy][cx]
                    if v == 9:
                        continue
                    if rotate:
                        gx = x + (len(m) - cy) + 1
                        gy = y + cx + 2
                    else:
                        gx = x + cx + 2
                        gy = y + cy + 2
                    pos = (self.width * gy) + gx
                    idx = pos * 4
                    a2 = (9.0 - v) / 9.0
                    r = (1.0 - a2) * self.buf[idx]
                    g = (1.0 - a2) * self.buf[idx + 1]
                    b = (1.0 - a2) * self.buf[idx + 2]
                    a1 = self.buf[idx + 3] / 255.0
                    a = a2 + a1 * (1.0 - a2)
                    try:
                        self.buf[idx] = int(r)
                        self.buf[idx + 1] = int(g)
                        self.buf[idx + 2] = int(b)
                        self.buf[idx + 3] = int(a * 255)
                    except IndexError:
                        pass
            if rotate:
                y += 11
            else:
                x += 11


def draw_guides(draw_buff, extends):
    width = int(extends[2] - extends[0])
    height = int(extends[3] - extends[1])
    draw_buff.set_color(0, 0, 0, 255)
    draw_buff.line_width = 1
    min_x = int(extends[0])
    min_y = int(extends[1])
    points = 50
    draw_buff.draw_text(0, 0, "mm")
    for x in range(points - (min_x % points), width - 30, points):
        if x < 30:
            continue
        draw_buff.draw_text(x, 0, str(int((x + min_x) / 10)), rotate=True)
        draw_buff.draw_line(x, 0, x, 30)
    for y in range(points - (min_y % points), height - 30, points):
        if y < 30:
            continue
        draw_buff.draw_text(0, y, str(int((y + min_y) / 10)))
        draw_buff.draw_line(0, y, 30, y)


def write(pattern, f, settings=None):
    guides = settings.get("guides", False)
    extends = pattern.bounds()
    pattern.translate(-extends[0], -extends[1])
    width = int(extends[2] - extends[0])
    height = int(extends[3] - extends[1])
    draw_buff = PngBuffer(width, height)
    draw_buff.fancy = settings.get("fancy", False)
    if settings is not None:
        background = settings.get("background")
        if background is not None:
            b = EmbThread()
            b.set(background)
            draw_buff.background(b.get_red(), b.get_green(), b.get_blue(), 0xFF)
        linewidth = settings.get("linewidth")
        if linewidth is not None and isinstance(linewidth, int):
            draw_buff.line_width = linewidth

    for stitchblock in pattern.get_as_stitchblock():
        block = stitchblock[0]
        thread = stitchblock[1]
        draw_buff.set_color(
            thread.get_red(), thread.get_green(), thread.get_blue(), 255
        )
        last_x = None
        last_y = None
        for stitch in block:
            x = int(stitch[0])
            y = int(stitch[1])
            if last_x is not None:
                draw_buff.draw_line(last_x, last_y, x, y)
            last_x = x
            last_y = y

    if guides:
        draw_guides(draw_buff, extends)

    f.write(write_png(draw_buff.buf, draw_buff.width, draw_buff.height))



from .EmbConstant import FAST, SLOW


def read_barudan_dat(f, out):
    count = 0
    while True:
        count += 1
        byte = bytearray(f.read(3))
        if len(byte) != 3:
            break
        ctrl = byte[0]
        dy = -byte[1]
        dx = byte[2]
        if ctrl & 0x80 == 0:
            # This bit should always be set, must be other dat type.
            return False
        if (ctrl & 0x20) != 0:
            dx = -dx
        if (ctrl & 0x40) != 0:
            dy = -dy
        command = ctrl & 0b11111
        if command == 0x0:
            # Stitch
            out.stitch(dx, dy)
            continue
        if command == 0x01:
            # Jump
            out.move(dx, dy)
            continue
        if command == 0x02:
            # Fast
            out.add_stitch_relative(FAST)
            if dx != 0 or dy != 0:
                out.stitch(dx, dy)
            continue
        if command == 0x03:
            # Fast, Jump
            out.add_stitch_relative(FAST)
            if dx != 0 or dy != 0:
                out.move(dx, dy)
            continue
        if command == 0x04:
            # Slow
            out.add_stitch_relative(SLOW)
            if dx != 0 or dy != 0:
                out.stitch(dx, dy)
            continue
        if command == 0x05:
            # Slow, Jump
            out.add_stitch_relative(SLOW)
            if dx != 0 or dy != 0:
                out.move(dx, dy)
            continue
        if command == 0x06:
            # T1 Top Thread Trimming, TTrim.
            out.trim()
            if dx != 0 or dy != 0:
                out.move(dx, dy)
            continue
        if command == 0x07:
            # T2 Bobbin Threading
            out.trim()
            if dx != 0 or dy != 0:
                out.move(dx, dy)
            continue
        if (
            command == 0x08
        ):  # ww, stop file had proper A8 rather than E8 and displacement
            # C00 Stop
            out.stop()
            if dx != 0 or dy != 0:
                out.move(dx, dy)
            continue
        if 0x09 <= command <= 0x17:
            # C01 - C14
            needle = command - 0x08
            out.needle_out(needle)
            if dx != 0 or dy != 0:
                out.move(dx, dy)
            continue
        if command == 0x18:
            break
        if ctrl == 0x2B:
            break  # Rare postfix data from machine. Do not read this.
        break  # Uncaught Command
    out.end()
    return True


def read_sunstar_dat_stitches(f, out):
    count = 0
    while True:
        count += 1
        byte = bytearray(f.read(3))
        if len(byte) != 3:
            break
        x = byte[0] & 0x7F
        y = byte[1] & 0x7F
        if byte[0] & 0x80:
            x = -x
        if byte[1] & 0x80:
            y = -y
        y = -y
        ctrl = byte[2]
        if ctrl == 0x07:
            out.stitch(x, y)
            continue
        if ctrl == 0x04:
            out.move(x, y)
            continue
        if ctrl == 0x80:
            out.trim()
            if x != 0 or y != 0:
                out.move(x, y)
            continue
        if ctrl == 0x87:
            out.color_change()
            if x != 0 or y != 0:
                out.move(x, y)
            continue
        if ctrl == 0x84:  # Initialized info.
            out.stitch(x, y)
            continue
        elif ctrl == 0:
            break
        break  # Uncaught Control
    out.end()


def read_sunstar_dat(f, out):
    # f.seek(0x02, 0)
    # stitches = read_int_16le(f)
    f.seek(0x100, 0)
    read_sunstar_dat_stitches(f, out)


def read(f, out, settings=None):
    if not read_barudan_dat(f, out):
        f.seek(0, 0)
        read_sunstar_dat(f, out)



import datetime

from .EmbConstant import *
from .EmbThreadJef import get_thread_set
from .WriteHelper import write_int_8, write_int_32le, write_string_utf8

SEQUIN_CONTINGENCY = CONTINGENCY_SEQUIN_JUMP
FULL_JUMP = True
ROUND = True
MAX_JUMP_DISTANCE = 127
MAX_STITCH_DISTANCE = 127

# These are in mm, embroidery units are 1/10 mm
HOOP_110X110 = 0
HOOP_50X50 = 1
HOOP_140X200 = 2
HOOP_126X110 = 3
HOOP_200X200 = 4


def write(pattern, f, settings=None):
    trims = False
    command_count_max = 3

    date_string = datetime.datetime.today().strftime("%Y%m%d%H%M%S")
    if settings is not None:
        trims = settings.get("trims", trims)
        command_count_max = settings.get("trim_at", command_count_max)
        date_string = settings.get("date", date_string)

    pattern.fix_color_count()
    # REMOVE BUG: color_count = pattern.count_threads(). #

    # PATCH
    jef_threads = get_thread_set()
    last_index = None
    last_thread = None
    palette = []
    color_toggled = False
    color_count = 0  # Color and Stop count.
    index_in_threadlist = 0
    for stitch in pattern.stitches:
        # Iterate all stitches.
        flags = stitch[2] & COMMAND_MASK
        if flags == COLOR_CHANGE or index_in_threadlist == 0:
            # If color change *or* initial color unset.
            thread = pattern.threadlist[index_in_threadlist]
            index_in_threadlist += 1
            color_count += 1
            index_of_jefthread = thread.find_nearest_color_index(jef_threads)
            if last_index == index_of_jefthread and last_thread != thread:
                # Last thread and current thread pigeonhole to same jefcolor.
                # We set that thread to None. And get the second closest color.
                repeated_thread = jef_threads[index_of_jefthread]
                repeated_index = index_of_jefthread
                jef_threads[index_of_jefthread] = None
                index_of_jefthread = thread.find_nearest_color_index(jef_threads)
                jef_threads[repeated_index] = repeated_thread
            palette.append(index_of_jefthread)
            last_index = index_of_jefthread
            last_thread = thread
            color_toggled = False
        if flags == STOP:
            color_count += 1
            color_toggled = not color_toggled
            if color_toggled:
                palette.append(0)
            else:
                palette.append(last_index)
    # END PATCH

    offsets = 0x74 + (color_count * 8)
    write_int_32le(f, offsets)
    write_int_32le(f, 0x14)
    write_string_utf8(f, date_string)
    write_int_8(f, 0)
    write_int_8(f, 0)
    write_int_32le(f, color_count)
    point_count = 1  # 1 command for END statement
    for stitch in pattern.stitches:
        data = stitch[2] & COMMAND_MASK
        if data == STITCH:
            point_count += 1
        elif data == JUMP:
            point_count += 2
        elif data == TRIM:
            if trims:
                point_count += 2 * command_count_max
        elif data == COLOR_CHANGE or data == STOP:  # PATCH: INCLUDE STOP.
            point_count += 2
        elif data == END:
            break
    write_int_32le(f, point_count)
    extends = pattern.bounds()
    design_width = int(round(extends[2] - extends[0]))
    design_height = int(round(extends[3] - extends[1]))
    write_int_32le(f, get_jef_hoop_size(design_width, design_height))
    half_width = int(round(design_width / 2))
    half_height = int(round(design_height / 2))

    # distance from center of hoop.
    write_int_32le(f, half_width)
    write_int_32le(f, half_height)
    write_int_32le(f, half_width)
    write_int_32le(f, half_height)

    # distance from default 110 x 110 hoop
    x_hoop_edge = 550 - half_width
    y_hoop_edge = 550 - half_height
    write_hoop_edge_distance(f, x_hoop_edge, y_hoop_edge)

    # distance from default 50 x 50 hoop
    x_hoop_edge = 250 - half_width
    y_hoop_edge = 250 - half_height
    write_hoop_edge_distance(f, x_hoop_edge, y_hoop_edge)

    # distance from default 140 x 200 hoop
    x_hoop_edge = 700 - half_width
    y_hoop_edge = 1000 - half_height
    write_hoop_edge_distance(f, x_hoop_edge, y_hoop_edge)

    # distance from custom hoop, but this should be accepted.
    x_hoop_edge = 700 - half_width
    y_hoop_edge = 1000 - half_height
    write_hoop_edge_distance(f, x_hoop_edge, y_hoop_edge)

    # REMOVE (We covered this in PATCH).
    #jef_threads = get_thread_set()
    #
    #palette = build_nonrepeat_palette(jef_threads, pattern.threadlist)
    # END REMOVE

    for t in palette:
        write_int_32le(f, t)

    for i in range(0, color_count):
        write_int_32le(f, 0x0D)

    xx = 0
    yy = 0
    for stitch in pattern.stitches:
        x = stitch[0]
        y = stitch[1]
        data = stitch[2] & COMMAND_MASK
        dx = int(round(x - xx))
        dy = int(round(y - yy))
        xx += dx
        yy += dy
        if data == STITCH:
            write_int_8(f, dx)
            write_int_8(f, -dy)
            continue
        elif data == COLOR_CHANGE or data == STOP:  # PATCH INCLUDE STOP.
            f.write(b"\x80\x01")
            write_int_8(f, dx)
            write_int_8(f, -dy)
            continue
        elif data == TRIM:
            if trims:  # command trim.
                f.write(b"\x80\x02\x00\x00" * command_count_max)
            continue
        elif data == JUMP:
            f.write(b"\x80\x02")
            write_int_8(f, dx)
            write_int_8(f, -dy)
            continue
        elif data == END:
            break
    f.write(b"\x80\x10")


def get_jef_hoop_size(width, height):
    if width < 500 and height < 500:
        return HOOP_50X50
    if width < 1260 and height < 1100:
        return HOOP_126X110
    if width < 1400 and height < 2000:
        return HOOP_140X200
    if width < 2000 and height < 2000:
        return HOOP_200X200
    return HOOP_110X110


def write_hoop_edge_distance(f, x_hoop_edge, y_hoop_edge):
    if min(x_hoop_edge, y_hoop_edge) >= 0:
        write_int_32le(f, x_hoop_edge)  # left
        write_int_32le(f, y_hoop_edge)  # top
        write_int_32le(f, x_hoop_edge)  # right
        write_int_32le(f, y_hoop_edge)  # bottom
    else:
        write_int_32le(f, -1)
        write_int_32le(f, -1)
        write_int_32le(f, -1)
        write_int_32le(f, -1)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyembroidery",
    version="1.5.1",
    author="Tatarize",
    author_email="tatarize@gmail.com",
    description="Embroidery IO library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/EmbroidePy/pyembroidery",
    packages=setuptools.find_packages(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
)


from __future__ import print_function

import unittest

from test.pattern_for_tests import *
from pyembroidery import GenericWriter


class TestConverts(unittest.TestCase):
    def test_generic_write_stitch(self):
        file1 = "convert.dst"
        file2 = "convert.txt"
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)
        pattern = get_fractal_pattern()
        pattern.write(file1)

        pattern = EmbPattern(file1)

        EmbPattern.write_embroidery(
            GenericWriter,
            pattern,
            file2,
            {
                "segment_start": "\t\t",
                "segment_end": "\n",
                "segment": "{cmd_str} {x},{y}",
                "stitch": "Stitch: {x},{y}",
                "jump": "Jump {x},{y}",
                "trim": "Trim: {x},{y}",
                "color_change": "Color-Change: {x},{y}",
                "block_start": "\t{{\n",
                "block_end": "\t}}\n",
                "color_start": "[\n",
                "color_end": "]\n",
            },
        )
        print("write generic: ", file1)

    def test_generic_write_gcode(self):
        gcode_writer_dict = {
            "scale": (-0.1, -0.1),
            "pattern_start": "(STITCH_COUNT: {stitch_total})\n"
            "(THREAD_COUNT: {color_change_count})\n"
            "(EXTENTS_LEFT: {extents_left:.3f})\n"
            "(EXTENTS_TOP: {extends_top:.3f})\n"
            "(EXTENTS_RIGHT: {extends_right:.3f})\n"
            "(EXTENTS_BOTTOM: {extends_bottom:.3f})\n"
            "(EXTENTS_WIDTH: {extends_width:.3f})\n"
            "(EXTENTS_HEIGHT: {extends_height:.3f})\n"
            "(COMMAND_STITCH: {stitch_count})\n"
            "(COMMAND_COLOR_CHANGE: {color_change_count})\n"
            "(COMMAND_TRIM: {trim_count})\n"
            "(COMMAND_STOP: {stop_count})\n"
            "(COMMAND_END: {end_count})\n",
            "metadata_entry": "({metadata_key}: {metadata_value})\n",
            "thread_entry": "(Thread{thread_index}: {thread_color} {thread_description} {thread_brand} {thread_catalog_number})\n",
            "stitch": "G00 X{x:.3f} Y{y:.3f}\nG00 Z{z:.1f}\n",
            "color_change": "M00\n",
            "stop": "M00\n",
            "end": "M30\n",
        }
        file1 = "file-gcode.gcode"
        file2 = "file-generic.gcode"
        pattern = get_fractal_pattern()
        pattern.fix_color_count()
        EmbPattern.write_embroidery(
            GenericWriter,
            pattern,
            file2,
            gcode_writer_dict,
        )
        pattern.write(file1)
        f1 = open(file1, "rb").read()
        f2 = open(file2, "rb").read()
        self.assertEqual(f1, f2)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)



from .ReadHelper import read_int_32le, read_int_16le, read_int_8, read_int_24be, signed8


def read_zhs_stitches(f, out):
    count = 0

    xx = 0
    yy = 0
    while True:
        count += 1
        b = bytearray(f.read(3))
        if len(b) != 3:
            break
        ctrl = b[0]
        if ctrl == 0x10:
            # Checksum
            continue
        # x
        x = 0
        x += b[1] & 0b00000001
        x += b[2] & 0b00000010
        x += b[1] & 0b00000100
        x += b[2] & 0b00001000
        x += b[1] & 0b00010000
        x += b[2] & 0b00100000
        x += b[1] & 0b01000000
        x += b[2] & 0b10000000
        x = signed8(x)
        if x >= 63:
            x += 1
        if x <= -63:
            x -= 1

        # y
        y = 0
        y += b[2] & 0b00000001
        y += b[1] & 0b00000010
        y += b[2] & 0b00000100
        y += b[1] & 0b00001000
        y += b[2] & 0b00010000
        y += b[1] & 0b00100000
        y += b[2] & 0b01000000
        y += b[1] & 0b10000000

        y = signed8(y)
        if y >= 63:
            y += 1
        if y <= -63:
            y -= 1

        xx += x
        yy += y
        if ctrl == 0x41:
            # Still unmapped.
            pass
        elif ctrl == 0x02:
            out.stitch(xx, -yy)
            xx = 0
            yy = 0
            continue
        elif ctrl == 0x01:
            out.move(xx, -yy)
            xx = 0
            yy = 0
            continue
        elif ctrl == 0x04:
            xx = 0
            yy = 0
            out.color_change()
            continue
        elif ctrl == 0x80:
            break
    out.end()


def read_zhs_header(f, out):
    color_count = read_int_8(f)
    for i in range(color_count):
        out.add_thread(read_int_24be(f))
    length = read_int_16le(f)
    b = bytearray(f.read(length))
    thread_data = b.decode('utf8')
    threads = thread_data.split("&$")
    try:
        for i, data in enumerate(threads[1:]):
            thread = out.threadlist[i]
            parts = data.split("&#")
            try:
                if len(parts[0]):
                    thread.chart = parts[0]
                if len(parts[1]):
                    thread.description = parts[1]
                if len(parts[2]) > 3:
                    thread.catalog_number = parts[2][:-2]
            except IndexError:
                pass
    except IndexError:
        pass


def read(f, out, settings=None):
    f.seek(0x0F, 0)
    stitch_start_position = read_int_32le(f)
    header_start_position = read_int_32le(f)
    f.seek(header_start_position, 0)
    read_zhs_header(f, out)
    f.seek(stitch_start_position, 0)
    read_zhs_stitches(f, out)
